static decl SKDE_BANDWIDTH = 0.5;
static decl SP_MAX_ITERATIONS = 5;

/*
	The GAS(1,1) filters.

	Common arguments:
		data: a 1xT vector of data
		mu: mean
		omega, alpha, beta: GAS parameters

	Returns an array with the following elements:
	[0]: filtered f_t
	[1]: filtered standardized residuals
*/

GASfilter_Normal(const data, const mu, const omega, const alpha, const beta) {
	decl T = columns(data), f_t = zeros(1, T), eps_t = zeros(1, T);
	f_t[0] = omega;
	decl i = 0;
	while (TRUE) {
		eps_t[i] = (data[i] - mu) * exp(-f_t[i]/2);
		if(i == T-1) break;
		f_t[i+1] = omega * (1 - beta) + alpha * (-1/2) * (eps_t[i] * dlogn(eps_t[i]) + 1) + beta * f_t[i];
		++i;
	}
	return {f_t, eps_t};
}

// extra argument: nu
GASfilter_t(const data, const mu, const omega, const alpha, const beta, const nu) {
	decl T = columns(data), f_t = zeros(1, T), eps_t = zeros(1, T);
	f_t[0] = omega;
	decl i = 0;
	while (TRUE) {
		eps_t[i] = (data[i] - mu) * exp(-f_t[i]/2);
		if(i == T-1) break;
		f_t[i+1] = omega * (1 - beta) + alpha * (-1/2) * (eps_t[i] * dlognormt(eps_t[i], nu) + 1) + beta * f_t[i];
		++i;
	}
	return {f_t, eps_t};
}

// extra argument: StandardizedKernelEstimator object
GASfilter_SP(const data, const mu, const omega, const alpha, const beta, const skde) {
	decl T = columns(data), f_t = zeros(1, T), eps_t = zeros(1, T);
	f_t[0] = omega;
	decl i = 0;
	while (TRUE) {
		eps_t[i] = (data[i] - mu) * exp(-f_t[i]/2);
		if(i == T-1) break;
		f_t[i+1] = omega * (1 - beta) + alpha * (-1/2) * (eps_t[i] * skde.dlog(eps_t[i]) + 1) + beta * f_t[i];
		++i;
	}
	return {f_t, eps_t};
}

GASfilter_MN(const data, const mu, const omega, const alpha, const beta) {
	decl T = columns(data), f_t = zeros(1, T), eps_t = zeros(1, T);
	f_t[0] = omega;
	decl i = 0;
	while (TRUE) {
		eps_t[i] = (data[i] - mu) * exp(-f_t[i]/2);
		if(i == T-1) break;
		f_t[i+1] = omega * (1 - beta) + alpha * (-1/2) * (eps_t[i] * dlogchimix(eps_t[i]) + 1) + beta * f_t[i];
		++i;
	}
	return {f_t, eps_t};
}

/*
	The GARCH(1,1) filter. Returns filtered sigma^2_t instead of f_t.
*/
GARCHfilter(const data, const mu, const omega, const alpha, const beta) {
	decl T = columns(data), sigma2_t = zeros(1, T), xi_t = zeros(1, T);
	sigma2_t[0] = omega;
	decl i = 0;
	while (TRUE) {
		xi_t[i] = data[i] - mu;
		if(i == T-1) break;
		sigma2_t[i+1] = omega * (1 - beta - alpha) + alpha * xi_t[i]^2 + beta * sigma2_t[i];
		++i;
	}
	return {sigma2_t, xi_t./sqrt(sigma2_t)};
}

/*
	Estimators. All estimators have the following form:

	Arguments:
		data: a 1xT vector of data
		K: how much of the data to use for estimation
		initpar: starting parameters for optimization. A row vector with elements mu, omega, alpha, beta,
			and possibly nu for estimators of models with t distributions

	Returns a dictionary (Dict) with all results.
		This dictionary will always at least have a key "succes", which indicates if optimization
		was succesful or not.
*/

/*
	First: specific GAS MLE's.
*/

GAS_Normal(const data, const K, const initpar) {
	print("N");
	decl transpar = initpar';
	decl val;

	decl code = MaxBFGS(
		[=] (transpar, const aval, const a2, const a3) {
			decl par = transpar';
			decl mu = par[0], omega = par[1], alpha = par[2], beta = par[3];
			
			decl f_t, eps_t;
			[f_t, eps_t] = GASfilter_Normal(data[:K-1], mu, omega, alpha, beta);
			aval[0] = meanr(log(densn(eps_t))) - meanr(0.5 * f_t);
			
			return 1;
		}
	, &transpar, &val, 0, 1);

	if(code != 0) return new Dict({"succes", FALSE});

	decl par = transpar';
	decl mu = par[0], omega = par[1], alpha = par[2], beta = par[3];
	decl f_t, eps_t;
	[f_t, eps_t] = GASfilter_Normal(data, mu, omega, alpha, beta);
	
	return new Dict({"succes", TRUE, "par", par, "loglik", val, "ft", f_t, "epst", eps_t});	
}

GAS_t3(const data, const K, const initpar) {
	print("t3");
	decl transpar = initpar';
	decl val;

	decl code = MaxBFGS(
		[=] (transpar, const aval, const a2, const a3) {
			decl par = transpar';
			decl mu = par[0], omega = par[1], alpha = par[2], beta = par[3];
			
			decl f_t, eps_t;
			[f_t, eps_t] = GASfilter_t(data[:K-1], mu, omega, alpha, beta, 3);
			aval[0] = meanr(log(densnormt(eps_t, 3))) - meanr(0.5 * f_t);
			
			return 1;
		}
	, &transpar, &val, 0, 1);

	if(code != 0) return new Dict({"succes", FALSE});

	decl par = transpar';
	decl mu = par[0], omega = par[1], alpha = par[2], beta = par[3];
	decl f_t, eps_t;
	[f_t, eps_t] = GASfilter_t(data, mu, omega, alpha, beta, 3);
	
	return new Dict({"succes", TRUE, "par", par, "loglik", val, "ft", f_t, "epst", eps_t});	
}

GAS_t5(const data, const K, const initpar) {
	print("t5");
	decl transpar = initpar';
	decl val;

	decl code = MaxBFGS(
		[=] (transpar, const aval, const a2, const a3) {
			decl par = transpar';
			decl mu = par[0], omega = par[1], alpha = par[2], beta = par[3];
			
			decl f_t, eps_t;
			[f_t, eps_t] = GASfilter_t(data[:K-1], mu, omega, alpha, beta, 5);
			aval[0] = meanr(log(densnormt(eps_t, 5))) - meanr(0.5 * f_t);
			
			return 1;
		}
	, &transpar, &val, 0, 1);

	if(code != 0) return new Dict({"succes", FALSE});

	decl par = transpar';
	decl mu = par[0], omega = par[1], alpha = par[2], beta = par[3];
	decl f_t, eps_t;
	[f_t, eps_t] = GASfilter_t(data, mu, omega, alpha, beta, 5);
	
	return new Dict({"succes", TRUE, "par", par, "loglik", val, "ft", f_t, "epst", eps_t});	
}

GAS_MN(const data, const K, const initpar) {
	print("MN");
	decl transpar = initpar';
	decl val;

	decl code = MaxBFGS(
		[=] (transpar, const aval, const a2, const a3) {
			decl par = transpar';
			decl mu = par[0], omega = par[1], alpha = par[2], beta = par[3];
			
			decl f_t, eps_t;
			[f_t, eps_t] = GASfilter_MN(data[:K-1], mu, omega, alpha, beta);
			aval[0] = meanr(log(denschimix(eps_t))) - meanr(0.5 * f_t);
			
			return 1;
		}
	, &transpar, &val, 0, 1);

	if(code != 0) return new Dict({"succes", FALSE});

	decl par = transpar';
	decl mu = par[0], omega = par[1], alpha = par[2], beta = par[3];
	decl f_t, eps_t;
	[f_t, eps_t] = GASfilter_MN(data, mu, omega, alpha, beta);
	
	return new Dict({"succes", TRUE, "par", par, "loglik", val, "ft", f_t, "epst", eps_t});	
}

/*
	t-GAS with arbitrary degrees of freedom
*/

GAS_t(const data, const K, const initpar) {
	print("t");
	decl transpar = Transform(initpar, <0, 0, 0, 0, 1>)';
	decl val;

	decl code = MaxBFGS(
		[=] (transpar, const aval, const a2, const a3) {
			decl par = Transform(transpar, <0, 0, 0, 0, -1>)';
			decl mu = par[0], omega = par[1], alpha = par[2], beta = par[3], nu = par[4];
			
			decl f_t, eps_t;
			[f_t, eps_t] = GASfilter_t(data[:K-1], mu, omega, alpha, beta, nu);
			aval[0] = meanr(log(densnormt(eps_t, nu))) - meanr(0.5 * f_t);
			
			return 1;
		}
	, &transpar, &val, 0, 1);

	if(code != 0) return new Dict({"succes", FALSE});

	decl par = Transform(transpar, <0, 0, 0, 0, -1>)';
	decl mu = par[0], omega = par[1], alpha = par[2], beta = par[3], nu = par[4];
	decl f_t, eps_t;
	[f_t, eps_t] = GASfilter_t(data, mu, omega, alpha, beta, nu);
	
	return new Dict({"succes", TRUE, "par", par, "loglik", val, "ft", f_t, "epst", eps_t});	
}

/*
	The collection of GAS models: QML, t, SP(1) and SP(AIC)).
*/

GAS(const data, const K, const initpar) {
	decl QMLresults = GAS_Normal(data, K, initpar[0:3]);
	if(!QMLresults.get("succes")) return QMLresults;

	decl Tresults = GAS_t(data, K, initpar);
	if(!Tresults.get("succes")) return Tresults;

	// begin SP(1)
	print("SP");
	
	decl skde = new StandardizedKernelEstimator(Tresults.get("epst")[:K-1], SKDE_BANDWIDTH);

	// use t estimate
	decl transpar = (Tresults.get("par")[0:3])';
	decl val;

	decl code = MaxBFGS(
		[=] (transpar, const aval, const a2, const a3) {
			decl par = transpar';
			decl mu = par[0], omega = par[1], alpha = par[2], beta = par[3];
			
			decl f_t, eps_t;
			[f_t, eps_t] = GASfilter_SP(data[:K-1], mu, omega, alpha, beta, skde);
			aval[0] = meanr(log(skde.dens(eps_t))) - meanr(0.5 * f_t);
			
			return 1;
		}
	, &transpar, &val, 0, 1);

	if(code != 0) return new Dict({"succes", FALSE});

	decl par = transpar';
	decl mu = par[0], omega = par[1], alpha = par[2], beta = par[3];
	decl f_t, eps_t;
	[f_t, eps_t] = GASfilter_SP(data, mu, omega, alpha, beta, skde);

	delete skde;

	// begin SP(AIC) selection

	decl SPiter = 1;
	// save SP(1)
	decl SP1val = val, SP1par = par, SP1f_t = f_t, SP1eps_t = eps_t;

	decl newval; 
	
	while(SPiter < SP_MAX_ITERATIONS) {
		// use previous transpar
		decl skde = new StandardizedKernelEstimator(eps_t[:K-1], SKDE_BANDWIDTH);

		decl code = MaxBFGS(
			[=] (transpar, const aval, const a2, const a3) {
				decl par = transpar';
				decl mu = par[0], omega = par[1], alpha = par[2], beta = par[3];
				
				decl f_t, eps_t;
				[f_t, eps_t] = GASfilter_SP(data[:K-1], mu, omega, alpha, beta, skde);
				aval[0] = meanr(log(skde.dens(eps_t))) - meanr(0.5 * f_t);
				
				return 1;
			}
		, &transpar, &newval, 0, 1);

		if(code != 0) return new Dict({"succes", FALSE});

		if(newval > val) {
			val = newval;
			par = transpar';
			decl mu = par[0], omega = par[1], alpha = par[2], beta = par[3];
			[f_t, eps_t] = GASfilter_SP(data, mu, omega, alpha, beta, skde);
			SPiter++;
		} else {
			break;
		}

		delete skde;
	}

	decl results = new Dict({
		"succes", TRUE,
		"QML-loglik", QMLresults.get("loglik"),
		"QML-par", QMLresults.get("par"),
		"QML-ft", QMLresults.get("ft"),
		"QML-epst", QMLresults.get("epst"),
		"t-loglik", Tresults.get("loglik"),
		"t-par", Tresults.get("par"),
		"t-ft", Tresults.get("ft"),
		"t-epst", Tresults.get("epst"),
		"SP1-loglik", SP1val,
		"SP1-par", SP1par,
		"SP1-ft", SP1f_t,
		"SP1-epst", SP1eps_t,
		"SPX-loglik", val,
		"SPX-par", par,
		"SPX-ft", f_t,
		"SPX-epst", eps_t,
		"SPiter", SPiter
	});

	delete QMLresults, Tresults;
	return results;
}

/*
	The GARCH equivalents of the GAS models.
*/

GARCH(const data, const K, const initpar) {

	// Normal
	
	decl transpar = Transform(initpar[0:3], <0, 1, 0, 2>)';
	decl normalval;

	decl code = MaxBFGS(
		[=] (transpar, const aval, const a2, const a3) {
			decl par = Transform(transpar, <0, -1, 0, -2>)';
			decl mu = par[0], omega = par[1], alpha = par[2], beta = par[3];
			
			decl sigma2_t, eps_t;
			[sigma2_t, eps_t] = GARCHfilter(data[:K-1], mu, omega, alpha, beta);
			aval[0] = meanr(log(densn(eps_t))) - meanr(0.5 * log(sigma2_t));
			
			return 1;
		}
	, &transpar, &normalval, 0, 1);

	if(code != 0) return new Dict({"succes", FALSE});

	decl normalpar = Transform(transpar, <0, -1, 0, -2>)';
	decl mu = normalpar[0], omega = normalpar[1], alpha = normalpar[2], beta = normalpar[3];
	decl normalsigma2_t, normaleps_t;
	[normalsigma2_t, normaleps_t] = GARCHfilter(data, mu, omega, alpha, beta);

	// t

	transpar = Transform(initpar, <0, 1, 0, 2, 1>)';
	decl tval;

	code = MaxBFGS(
		[=] (transpar, const aval, const a2, const a3) {
			decl par = Transform(transpar, <0, -1, 0, -2, -1>)';
			decl mu = par[0], omega = par[1], alpha = par[2], beta = par[3], nu = par[4];
			
			decl sigma2_t, eps_t;
			[sigma2_t, eps_t] = GARCHfilter(data[:K-1], mu, omega, alpha, beta);
			aval[0] = meanr(log(densnormt(eps_t, nu))) - meanr(0.5 * log(sigma2_t));
			
			return 1;
		}
	, &transpar, &tval, 0, 1);

	if(code != 0) return new Dict({"succes", FALSE});

	decl tpar = Transform(transpar, <0, -1, 0, -2, -1>)';
	mu = tpar[0], omega = tpar[1], alpha = tpar[2], beta = tpar[3];
	decl tsigma2_t, teps_t;
	[tsigma2_t, teps_t] = GARCHfilter(data, mu, omega, alpha, beta);

	// SP(1)

	decl skde = new StandardizedKernelEstimator(teps_t[:K-1], SKDE_BANDWIDTH);

	transpar = transpar[0:3]; // use previous t estimate
	decl SP1val;

	code = MaxBFGS(
		[=] (transpar, const aval, const a2, const a3) {
			decl par = Transform(transpar, <0, -1, 0, -2>)';
			decl mu = par[0], omega = par[1], alpha = par[2], beta = par[3];
			
			decl sigma2_t, eps_t;
			[sigma2_t, eps_t] = GARCHfilter(data[:K-1], mu, omega, alpha, beta);
			aval[0] = meanr(log(skde.dens(eps_t))) - meanr(0.5 * log(sigma2_t));
			
			return 1;
		}
	, &transpar, &SP1val, 0, 1);

	if(code != 0) return new Dict({"succes", FALSE});

	decl SP1par = Transform(transpar, <0, -1, 0, -2>)';
	mu = SP1par[0], omega = SP1par[1], alpha = SP1par[2], beta = SP1par[3];
	decl SP1sigma2_t, SP1eps_t;
	[SP1sigma2_t, SP1eps_t] = GARCHfilter(data, mu, omega, alpha, beta);	

	delete skde;

	// begin SP(AIC) selection

	decl SPiter = 1;
	// initialize
	decl SPXval = SP1val, SPXpar = SP1par, SPXsigma2_t = SP1sigma2_t, SPXeps_t = SP1eps_t;

	decl newval; 
	
	while(SPiter < SP_MAX_ITERATIONS) {
		// use previous transpar
		decl skde = new StandardizedKernelEstimator(SP1eps_t[:K-1], SKDE_BANDWIDTH);

		decl code = MaxBFGS(
			[=] (transpar, const aval, const a2, const a3) {
				decl par = Transform(transpar, <0, -1, 0, -2>)';
				decl mu = par[0], omega = par[1], alpha = par[2], beta = par[3];
				
				decl sigma2_t, eps_t;
				[sigma2_t, eps_t] = GARCHfilter(data[:K-1], mu, omega, alpha, beta);
				aval[0] = meanr(log(skde.dens(eps_t))) - meanr(0.5 * log(sigma2_t));
				
				return 1;
			}
		, &transpar, &newval, 0, 1);

		if(code != 0) return new Dict({"succes", FALSE});

		if(newval > SPXval) {
			SPXval = newval;
			SPXpar = Transform(transpar, <0, -1, 0, -2>)';
			decl mu = SPXpar[0], omega = SPXpar[1], alpha = SPXpar[2], beta = SPXpar[3];
			[SPXsigma2_t, SPXeps_t] = GARCHfilter(data, mu, omega, alpha, beta);
			SPiter++;
		} else {
			break;
		}

		delete skde;
	}

	return new Dict({
		"succes", TRUE,
		"QML-loglik", normalval,
		"QML-par", normalpar,
		"QML-sigma2t", normalsigma2_t,
		"QML-epst", normaleps_t,
		"t-loglik", tval,
		"t-par", tpar,
		"t-sigma2t", tsigma2_t,
		"t-epst", teps_t,
		"SP1-loglik", SP1val,
		"SP1-par", SP1par,
		"SP1-sigma2t", SP1sigma2_t,
		"SP1-epst", SP1eps_t,
		"SPX-loglik", SPXval,
		"SPX-par", SPXpar,
		"SPX-sigma2t", SPXsigma2_t,
		"SPX-epst", SPXeps_t,
		"SPiter", SPiter
	});
}
